from typing import Any, List, Union

from pyairtable import Table, formulas as fx, Api

from bam_core import settings


class Airtable(object):

    def __init__(
        self,
        base_id: str = settings.AIRTABLE_BASE_ID,
        token: str = settings.AIRTABLE_TOKEN,
    ):
        self.base_id = base_id
        self.token = token
        self.api = Api(token)

    def get_table(self, table_name: str) -> Table:
        """
        Get a table object from the Airtable API
        :param table_name: The name of the table to get
        :return Table
        """
        return Table(self.token, self.base_id, table_name)

    def get_view(
        self, table_name: str, view_name: str, fields: List[str] = []
    ) -> Table:
        """
        Get a table object from the Airtable API
        :param table_name: The name of the table to get
        :return Table
        """
        return self.api.all(
            self.base_id, table_name, view=view_name, fields=fields
        )

    def get_view_count(
        self, table: str, view: str, fields: List[str] = [], unique=False
    ) -> int:
        """
        Get a table object from the Airtable API
        :param table_name: The name of the table to get
        :return Table
        """
        records = self.get_view(
            table_name=table, view_name=view, fields=fields
        )
        if unique and len(fields) and len(records):
            return len(set([r.get(fields[0]) for r in records]))
        return len(records)

    # core table objects

    @property
    def assistance_requests(self) -> Table:
        return self.get_table(settings.AIRTABLE_ASSISTANCE_REQUESTS_TABLE_NAME)

    @property
    def essential_goods(self) -> Table:
        return self.get_table(settings.AIRTABLE_ESSENTIAL_GOODS_TABLE_NAME)

    @property
    def volunteers(self) -> Table:
        return self.get_table(settings.AIRTABLE_VOLUNTEERS_TABLE_NAME)

    @property
    def mesh_requests(self):
        return self.get_view(settings.AIRTABLE_ASSISTANCE_REQUESTS_TABLE_NAME, settings.AIRTABLE_MESH_VIEW_NAME)


    def filter_table(
        self,
        table: Union[str, Table],
        expressions: List[Any],
        match_any: bool = False,
        sort: List[str] = ["-Date Submitted"],
        fields=[],
    ):
        if isinstance(table, str):
            table = self.get_table(table)
        if not match_any:
            formula = fx.AND(*expressions)
        else:
            formula = fx.OR(*expressions)
        return table.all(formula=formula, sort=sort, fields=fields)



import os
import shutil
import tempfile
import time
from urllib.parse import urljoin
import logging
from typing import Any, Dict, Optional
import zipfile
import requests

from bam_core import settings
from bam_core.utils import serde

log = logging.getLogger(__name__)


class ListMonk(object):
    """
    See API documentation here: https://listmonk.app/docs/apis/apis/
    """

    def __init__(
        self,
        username=settings.LISTMONK_USERNAME,
        password=settings.LISTMONK_PASSWORD,
        base_url=settings.LISTMONK_BASE_URL,
    ):
        self.username = username
        self.password = password
        self.base_url = base_url
        self.session = requests.Session()
        self.session.auth = (self.username, self.password)

    def make_request(
        self,
        path: str,
        method: str,
        data: Optional[Dict[str, Any]] = None,
        params: Optional[Dict[str, Any]] = None,
        files: Optional[Dict[str, Any]] = None,
    ) -> requests.Response:
        """
        Make a request to the ListMonk API.
        """
        url = urljoin(self.base_url, path)
        if method == "GET":
            response = self.session.get(url, params=params, files=files)
        elif method == "POST":
            response = self.session.post(
                url, json=data, params=params, files=files
            )
        elif method == "PUT":
            response = self.session.put(
                url, json=data, params=params, files=files
            )
        elif method == "DELETE":
            response = self.session.delete(url, params=params, files=files)
        else:
            raise ValueError("Invalid method")
        return response

    def get_lists(self, **params):
        """
        Get all lists.
        """
        r = self.make_request("api/lists", "GET", params=params)
        return r.json()

    def get_list_by_name(self, name, **params):
        """
        Get a list by name.
        """
        lists = self.get_lists(**params)
        for l in lists.get("data", {}).get("results", []):
            if name.lower().strip() == l["name"].lower().strip():
                return l
        return None

    def start_import(self, params, file):
        """
        Import subscribers from a CSV file.
        """
        default_params = {
            "mode": "subscribe",  # subscribe or blocklist
            "delim": ",",  # delimiter in the uploaded file
            "lists": [1],  # array of list IDs to import into
            "overwrite": True,  # overwrite existing entries or skip them?
        }

        if isinstance(params["lists"][0], str):
            # lookup list ids using list names
            ids = []
            for name in params["lists"]:
                l = self.get_list_by_name(name)
                if not l:
                    raise ValueError(f"Invalid list name: {name}")
                ids.append(l["id"])
            params["lists"] = ids
        default_params.update(params)
        with open(file, "rb") as f:
            log.info(
                f"Importing subscribers from {file} using params: {default_params}"
            )
            file = f.read()
            r = self.make_request(
                "api/import/subscribers",
                "POST",
                params={"params": serde.obj_to_json(default_params)},
                files={"file": file},
            )
        return r.json()

    def get_import_status(self):
        """
        Get the status of an import.
        """
        r = self.make_request("api/import/subscribers", "GET")
        return r.json().get("data", {})

    def stop_imports(self):
        """
        Stop all imports.
        """
        r = self.make_request("api/import/subscribers", "DELETE")
        return r.json()

    def get_import_logs(self):
        """
        Get import logs.
        """
        r = self.make_request("api/import/subscribers/logs", "GET")
        return [l for l in r.json().get("data", "").split("\n") if l]

    def run_import(self, params, subscribers, **kwargs):
        """
        Run an import.
        """
        self.stop_imports()
        try:
            tempdir, file = self.write_subscriber_list_to_csv(
                subscribers, **kwargs
            )
            self.start_import(params, file)
            while True:
                status = self.get_import_status()
                log.info(f"got import status: {status}")
                logs = self.get_import_logs()
                for line in logs:
                    log.info(line)
                if status["status"] == "finished":
                    break
                elif status["status"] == "failed":
                    raise ValueError(f"Import failed: {logs}")
                elif status["status"] == "importing":
                    log.info("Waiting 10 seconds before checking status again")
                    time.sleep(10)
                    continue
        except Exception as e:
            self.stop_imports()
            log.error(e)
            raise e
        finally:
            shutil.rmtree(tempdir)

        return True

    @classmethod
    def write_subscriber_list_to_csv(
        cls, subscribers, name_field="First Name", email_field="Email"
    ):
        """
        Write a list of subscribers to a CSV file.
        """
        data = []
        for subscriber in subscribers:
            if "fields" in subscriber:
                subscriber = subscriber.get("fields", {})
            attributes = serde.obj_to_json(
                {
                    key: value
                    for key, value in subscriber.items()
                    if key not in [name_field, email_field]
                }
            )
            name = subscriber.get(name_field)
            data.append(
                {
                    "email": subscriber.get(email_field),
                    "name": name,
                    "attributes": attributes,
                }
            )
        tempdir = tempfile.mkdtemp()
        file = os.path.join(tempdir, "subscribers.csv")
        file_zip = os.path.join(tempdir, "subscribers.csv.zip")
        log.info(f"Writing {len(data)} subscribers to {file}")
        with open(file, "w") as f:
            f.write(serde.obj_to_csv(data))
        with zipfile.ZipFile(file_zip, mode="w") as f:
            f.write(file)
        return tempdir, file_zip



"""
Utilities for interacting with AWS S3 / Digital Ocean Spaces
"""
import os
from io import BytesIO
import tempfile
import logging
import urllib
from typing import Callable, List, Union, Optional

import boto3
import requests

from bam_core import settings
from bam_core.utils import etc

log = logging.getLogger(__name__)

BAM_STOR_DEFAULT_MIMETYPE = "binary/octet-stream"


def get_bucket_name_and_scheme(s3_url):
    """
    Get the bucket name and scheme from a s3 url
    """
    p = urllib.parse.urlparse(s3_url)
    if p.netloc == "":
        return p.scheme, p.path.split("/")[0]
    return p.scheme, p.netloc


def parse(s3_url) -> tuple:
    """
    Parse a s3 url into a bucket name and key
    :param s3_url: The s3 url
    :return tuple
    """
    scheme, bucket_name = get_bucket_name_and_scheme(s3_url)
    s3_prefix = f"{scheme}://{bucket_name}"
    if s3_url.startswith(s3_prefix):
        key = s3_url[len(s3_prefix) :]
    elif s3_url.startswith(bucket_name):
        key = s3_url[len(bucket_name) + 1 :]
    else:
        key = s3_url
    if key.startswith("/"):
        key = key[1:]
    return bucket_name, key


class S3(object):
    def __init__(
        self,
        bucket_name: str = settings.S3_BUCKET,
        aws_access_key_id: str = settings.S3_ACCESS_KEY_ID,
        aws_secret_access_key: str = settings.S3_SECRET_ACCESS_KEY,
        endpoint_url: str = settings.S3_ENDPOINT_URL,
        base_url: str = settings.S3_BASE_URL,
        region_name: Optional[str] = settings.S3_REGION_NAME,
        platform: str = settings.S3_PLATFORM,
    ):
        self.scheme, self.bucket_name = get_bucket_name_and_scheme(bucket_name)
        self.access_key = aws_access_key_id
        self.access_secret = aws_secret_access_key
        self.endpoint_url = endpoint_url
        self.base_url = base_url
        self.region_name = region_name
        self.platform = platform
        self.resource = self.connect_resource()
        self.client = self.connect_client()
        self.bucket = self.get_bucket()
        if not self.scheme:
            self.scheme = "s3"
        if self.platform == "s3":
            self.scheme = "s3"
        self.s3_prefix = f"{self.scheme}://{self.bucket_name}/"

    # ////////////////////////
    #  Absolute Key Formatting
    #  (Allow all keys to have full s3:// paths on input and force full paths on output)
    # ///////////////////////

    def _in_key(self, key: str) -> str:
        f"""
        Format input key to accept full paths.
        :param key: An S3 key
        :return str
        """
        if key.startswith(self.s3_prefix):
            key = key.replace(self.s3_prefix, "")
        return key

    def _out_key(self, key: str) -> str:
        f"""
        Format output key to return full paths.
        :param key: An S3 key
        :return str
        """
        if not key.startswith(self.s3_prefix):
            key = f"{self.s3_prefix}{key}"
        return key

    def ensure(self) -> bool:
        """
        Ensure this bucket exists, return True if does, False if we create it
        """
        response = self.client.list_buckets()
        for bucket in response.get("Buckets", []):
            if bucket["Name"] == self.bucket_name:
                return True
        self.client.create_bucket(Bucket=self.bucket_name)
        return False

    # ////////////////////////
    #  Boto Client / Resource Connections
    # ///////////////////////
    @property
    def connection_kwargs(self):
        return dict(
            region_name=self.region_name,
            aws_access_key_id=self.access_key,
            aws_secret_access_key=self.access_secret,
            endpoint_url=self.endpoint_url,
        )

    def connect_resource(self):
        """
        Connect to boto3 s3 resource
        """
        return boto3.resource("s3", **self.connection_kwargs)

    def connect_client(self):
        """
        Connect to boto3 s3 resource
        """
        return boto3.client("s3", **self.connection_kwargs)

    def get_bucket(self):
        """
        Get boto3 bucket object
        """
        return self.resource.Bucket(self.bucket_name)

    # ////////////////////////
    #  Core Methods
    # ///////////////////////

    def get_meta(self, key: str) -> dict:
        f"""
        Fetch metadata about this file on S3.
        :param key: An S3 key
        :return dict
        """
        return self.client.head_object(
            Bucket=self.bucket_name, Key=self._in_key(key)
        )

    def get_contents(self, key: set):
        f"""
        Fetch the file contents from a key.
        :param key: An S3 key
        :return dict
        """
        obj = self.resource.Object(self.bucket_name, self._in_key(key))
        return obj.get()["Body"].read()

    def exists(self, key: str) -> bool:
        f"""
        Check whether this key exists
        :param key: An S3 key
        :return bool
        """
        objs = list(self.bucket.objects.filter(Prefix=self._in_key(key)))
        if len(objs) > 0 and objs[0].key == self._in_key(key):
            return True
        return False

    def download(self, key: str, local_path: Union[None, str] = None) -> str:
        f"""
        Download an s3 key to a local file
        :param key: An S3 key
        :param local_path: The local filepath to write to. if it doesn't exist, the file will be written to a tempfile and the path will be outputted."
        :return str
        """
        if local_path is None:
            local_path = os.path.join(
                tempfile.mkdtemp(), os.path.basename(key)
            )
        self.bucket.download_file(self._in_key(key), local_path)
        return local_path

    def download_file_obj(
        self, key: str, fobj: Optional[BytesIO] = None
    ) -> BytesIO:
        f"""
        Download an s3 key to a file-like object
        :param key: An S3 key
        :param fobj: A file-like object to write to. If not provided, the function will create an `io.BytesIO` object, write the file contents to it, and return it.
        :return BytesIO
        """
        fobj = BytesIO()
        self.client.download_fileobj(self.bucket_name, self._in_key(key), fobj)
        return fobj

    def download_all(
        self,
        prefix: str,
        local_path: Union[None, str] = None,
        key_filter: Callable = lambda x: True,
    ):
        f"""
        Download s3 files under a given prefix to a local directory. Returns the list of local filepaths.
        :param prefix: A prefix used to identify a list of s3 keys
        :param local_path: The local filepath to write to. if it doesn't exist, the file will be written to a tempfile and the path will be outputted."
        :param key_filter: A function that accepts a key and returns true if we should include the key in the results
        :yield str
        """
        if local_path is None:
            local_path = tempfile.gettempdir(prefix="bam_")
        os.makedirs(local_path, exist_ok=True)

        for key in self.list_keys(prefix, key_filter):
            dl_path = os.path.join(local_path, os.path.basename(key))
            self.bucket.download_file(self._in_key(key), dl_path)
            yield dl_path

    def upload_file_obj(
        self, fobj: BytesIO, key: str, mimetype: Optional[str] = None
    ) -> None:
        f"""
        Upload a file object to s3, optionally setting its mimetype
        :param key: An S3 key
        :param fobj: A file-like object to write to. If not provided, the function will create an `io.BytesIO` object, write the file contents to it, and return it.
        :return None
        """
        self.bucket.upload_fileobj(
            fobj,
            self._in_key(key),
            ExtraArgs={
                "ContentType": mimetype or BAM_STOR_DEFAULT_MIMETYPE,
                "MetadataDirective": "REPLACE",
            },
        )

    def _upload_file(
        self, local_path: str, key: str, mimetype: BAM_STOR_DEFAULT_MIMETYPE
    ) -> str:
        f"""
        Upload a file to a s3 bucket, optionally applying a mimetype
        :param local_path: The local filepath to write to. if it doesn't exist, the file will be written to a tempfile and the path will be outputted."
        :param key: An S3 key
        :param mimetype: The mimetype to set for this key
        :return str
        """
        self.bucket.upload_file(
            local_path,
            self._in_key(key),
            ExtraArgs={"ContentType": mimetype},
        )
        return self._out_key(key)

    def upload(
        self, local_path: str, key: str, mimetype: BAM_STOR_DEFAULT_MIMETYPE
    ):
        f"""
        Upload a file to a s3 bucket
        :param local_path: The local filepath to write to. if it doesn't exist, the file will be written to a tempfile and the path will be outputted."
        :param key: An S3 key
        :param mimetype: The mimetype to set for this key
        :return str
        """
        # TODO: replace all these os calls with ``path```
        if os.path.isdir(local_path):
            log.debug(
                "[upload] found directory at {local_path}. The default is to recursively upload from here."
            )
            for filename in etc.list_files(local_path):
                sub_path = os.path.relpath(filename, start=local_path)
                file_key = os.path.join(key, sub_path)
                log.debug(f" [s3-upload] UPLOADING {filename} to {file_key}")
                self._upload_file(
                    filename,
                    file_key,
                )

        return self._upload_file(local_path, key, mimetype)

    def delete(self, key: str) -> None:
        f"""
        Delete a file from s3.
        :param key: An S3 key
        :return None
        """
        obj = self.bucket.Object(self._in_key(key))
        obj.delete()

    def move(self, old_key: str, new_key: str, copy: bool = False) -> str:
        f"""
        Move a file on s3
        :param old_key: the file's current location
        :param new_key: the file's new location
        :param copy: whether or not to leave current file where it is.
        :return str
        """
        new_obj = self.bucket.Object(self._in_key(new_key))
        new_obj.copy(
            {"Bucket": self.bucket_name, "Key": self._in_key(old_key)},
            ExtraArgs={"MetadataDirective": "REPLACE"},
        )

        if not copy:
            old_obj = self.bucket.Object(self._in_key(old_key))
            old_obj.delete()
        return self._out_key(old_key)

    def move_all(
        self, old_pfx: str, new_pfx: str, copy: bool = False
    ) -> List[str]:
        f"""
        Move files on s3 returning their new paths
        :param old_pfx: the files' current prefix
        :param new_key: the files' new prefix
        :param copy: whether or not to leave current files where they are.
        :return list
        """
        new_paths = []
        for old_obj in self.bucket.objects.filter(Prefix=self.in_key(old_pfx)):
            new_key = old_obj.key.replace(old_pfx, new_pfx, 1)
            new_obj = self.bucket.Object(self._in_key(new_key))
            new_obj.copy(
                {"Bucket": self.bucket_name, "Key": old_obj.key},
                ExtraArgs={"MetadataDirective": "REPLACE"},
            )
            # cleanup
            if not copy:
                old_obj.delete()
            new_paths.append(self._out_key(new_key))
        return new_paths

    def copy(self, old_key: str, new_key: str) -> None:
        """"""
        return self.move(
            self._in_key(old_key), self._in_key(new_key), copy=True
        )

    def copy_all(self, old_pfx: str, new_pfx: str) -> List[str]:
        """"""
        return self.move_all(
            self._in_key(old_pfx), self._in_key(new_pfx), copy=True
        )

    def list_keys(self, prefix: str, key_filter: Callable = lambda x: True):
        f"""
        List keys in S3 bucket.
        :param prefix: A prefix used to identify a list of s3 keys
        :param prefix: A function that accepts a key and returns true if we should include the key in the results
        :yield str
        """
        return (
            self._out_key(obj.key)
            for obj in self.bucket.objects.filter(Prefix=self._in_key(prefix))
            if key_filter(obj.key)
        )

    # ////////////////////////
    #  Public/Private Access
    # ///////////////////////

    def set_acl(
        self, key: str, acl: str, raise_on_missing: bool = False
    ) -> None:
        f"""
        Set the access control for a s3 key
        :param key: An S3 key
        :param acl: The ACL string (either ``private`` or ``public-read``)
        :param raise_on_missing: Whether or not to raise an error if the key does not exist
        :return None
        """
        if not self.exists(key):
            if raise_on_missing:
                raise ValueError(f"{key} does not exist in {self.s3_prefix}")
            return
        obj = self.bucket.Object(self._in_key(key))
        obj.Acl().put(ACL=acl)

    def set_private(self, key: str) -> None:
        f"""
        Make this file on s3 private
        :param key: An S3 key
        :return None
        """
        self.set_acl(key, "private")

    def set_public(self, key: str):
        f"""
        Make this file on s3 private
        :param key: An S3 key
        :return None
        """
        self.set_acl(key, "public-read")

    def get_presigned_url(self, key: str, expiration: int = 3600) -> str:
        f"""
        Create a presigned url for an s3 asset.
        :param key: An S3 key
        :param expiration: The number of seconds this url is valid for.
        :return str
        """
        return self.client.generate_presigned_url(
            "get_object",
            Params={"Bucket": self.bucket_name, "Key": self._in_key(key)},
            ExpiresIn=expiration,
        )

    def get_public_url(self, key: str) -> str:
        f"""
        Get the public url for an s3 asset.
        :param key: An S3 key
        :return str
        """
        return "{}/{}/{}".format(
            self.client.meta.endpoint_url, self.bucket_name, self._in_key(key)
        )

    def purge_cdn_cache(self, prefix=""):
        """
        curl -X DELETE -H "Content-Type: application/json" \
        -H "Authorization: Bearer $API_TOKEN" \
        -d '{"files": ["*"]}' \
        "https://api.digitalocean.com/v2/cdn/endpoints/<CDN_ENDPOINT_ID>/cache"
        """
        r = requests.delete(
            f"https://api.digitalocean.com/v2/cdn/endpoints/{settings.S3_CDN_ID}/cache",
            headers={"Authorization": "Bearer {}".format(settings.DO_TOKEN)},
            json={"files": [f"{prefix}*"]},
        )
        r.raise_for_status()
        return True



from bam_core.utils.serde import obj_to_json, json_to_obj

def test_obj_to_json():
    assert obj_to_json({'a': 1}) == '{"a":1}'

def test_json_to_obj():
    assert json_to_obj('{"a": 1}') == {'a': 1}



"""
All things related to serialization / deserialization
This module should not import from other utils
"""

import csv
import io
import gzip
import zlib
import json
import pickle
from uuid import UUID
from decimal import Decimal
from inspect import isgenerator
from collections import Counter
from datetime import datetime
from typing import Any, Dict, List, Union

import yaml


# ///////////////////
# CLASSES
# ///////////////////


class SmartJSONEncoder(json.JSONEncoder):
    """JSON encoder extending Flask's default encoder to handle many different types"""

    item_separator = ","
    key_separator = ":"

    def default(self, obj: Any) -> str:
        """Return a serializable for ``o``, or call the base implementation."""
        if isinstance(obj, bytes):
            return obj.decode("utf-8")
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, (datetime.date, datetime.datetime, datetime.time)):
            return obj.strftime("%Y-%m-%dT%H:%M:%S.%fZ")
        if isinstance(obj, UUID):
            return str(obj)
        if isinstance(obj, Decimal):
            return float(obj)
        if isinstance(obj, set):
            return list(obj)
        if isgenerator(obj):
            return list(obj)
        if isinstance(obj, Counter):
            return dict(obj)
        if self.refs and hasattr(obj, "to_ref"):
            return obj.to_ref()
        if hasattr(obj, "to_dict"):
            return obj.to_dict()
        if hasattr(obj, "to_json"):
            return obj.to_json()
        return json.JSONEncoder.default(self, obj)


# ///////////////////
# FUNCTIONS
# ///////////////////


def file_to_gz(infile, outfile=None):
    """
    gzip a file
    """
    if not outfile:
        outfile = infile + ".gz"
    with gzip.open(outfile, "wb") as f_gz:
        with open(infile, "rb") as f_norm:
            f_gz.write(f_norm.read())


def gz_to_file(infile, outfile):
    """
    un-gzip a file
    """
    if not outfile:
        outfile = infile.replace(".gz", "")
    with gzip.open(infile, "rb") as f_gz:
        with open(outfile, "wb") as f_norm:
            f_gz.write(f_norm.read())


def str_to_gz_fobj(s: str) -> io.BytesIO:
    """
    string > gzip fobj
    """
    out = io.BytesIO()
    with gzip.GzipFile(fileobj=out, mode="w") as f:
        f.write(s)
    return out


def str_to_gz(s: str) -> bytes:
    """
    string > gzip bytes
    """
    return str_to_gz_fobj(s).getvalue()


def gz_to_str_fobj(b: bytes) -> io.BytesIO:
    """
    gzip > string
    """
    fileobj = io.BytesIO(b)
    return gzip.GzipFile(fileobj=fileobj, mode="r")


def gz_to_str(b: bytes) -> str:
    """
    gzip > str
    """
    f = gz_to_str_fobj(b)
    s = f.read()
    f.close()
    return s


def json_to_obj(s: Any) -> object:
    """
    json string > obj
    """
    # check for existing objects
    if isinstance(s, (dict, list)):
        return s
    return json.loads(s)


def obj_to_json(o: object) -> str:
    """
    obj > json string
    """
    return SmartJSONEncoder().encode(o)


def jsongz_to_obj(b: bytes) -> object:
    """
    json.gz > obj
    """
    return json_to_obj(gz_to_str(b))


def obj_to_jsongz(o: object) -> bytes:
    """
    obj > json.gz
    """
    return str_to_gz(obj_to_json(o))


def obj_to_jsongz_fobj(o: object) -> io.BytesIO:
    """
    obj > json.gz BytesIO
    """
    return str_to_gz_fobj(obj_to_json(o))


def pickle_to_obj(s: str) -> object:
    """
    pickle > obj
    """
    return pickle.loads(s)


def obj_to_pickle(o: object) -> str:
    """
    obj > pickle
    """
    return pickle.dumps(o)


def picklegz_to_obj(b: bytes) -> object:
    """
    pickle.gz > obj
    """
    return pickle_to_obj(gz_to_str(b))


def obj_to_picklegz(o: object) -> bytes:
    """
    obj > pickle.gz
    """
    return str_to_gz(obj_to_pickle(o))


def obj_to_picklegz_fobj(o: object) -> io.BytesIO:
    """
    obj > pickle.gz
    """
    return str_to_gz_fobj(obj_to_pickle(o))


def str_to_zip(s: str) -> bytes:
    """
    string > zip
    """
    if not isinstance(s, bytes):
        s = s.encode("utf-8")
    return zlib.compress(s)


def zip_to_str(b: bytes) -> str:
    """
    zip > string
    """
    return zlib.decompress(b)


def obj_to_yaml(o: object) -> str:
    """
    obj > yaml string
    """
    return yaml.dumps(o)


def yaml_to_obj(s: str) -> object:
    """
    yaml string > obj
    """
    return yaml.safe_load(s)


def obj_to_csv(o: List[Dict[str, Any]]) -> str:
    """
    obj > csv string
    """
    fieldnames = list(o[0].keys())
    fileobj = io.StringIO()
    writer = csv.DictWriter(
        fileobj,
        fieldnames=fieldnames,
        quotechar='"',
        quoting=csv.QUOTE_ALL,
        lineterminator="\n",
    )
    writer.writeheader()
    writer.writerows(o)
    csv_string = fileobj.getvalue()
    return csv_string


def csv_to_obj(s: str) -> List[Dict[str, Any]]:
    """
    csv string > obj
    """
    fileobj = io.StringIO(s)
    reader = csv.DictReader(fileobj)
    return [row for row in reader]

SERIALIZERS = {
    "json.gz": obj_to_jsongz,
    "json": obj_to_json,
    "pickle": obj_to_pickle,
    "pickle.gz": obj_to_picklegz,
    "zip": str_to_zip,
    "yaml": obj_to_yaml,
}

DESERIALIZERS = {
    "json.gz": jsongz_to_obj,
    "json": json_to_obj,
    "pickle": pickle_to_obj,
    "pickle.gz": picklegz_to_obj,
    "zip": zip_to_str,
    "yaml": yaml_to_obj,
}


def loads(s: Union[str, bytes], codec: str) -> object:
    """
    Serialize a string into a python object
    param s: a string or bytearray to load
    """
    if not codec in SERIALIZERS:
        raise NotImplementedError(f"[loads] Codec {codec} not supported")
    return SERIALIZERS.get(codec)(s)


def dumps(o: object, codec: str) -> Union[str, bytes]:
    """
    Serialize a string into a python object
    """
    if not codec in DESERIALIZERS:
        raise NotImplementedError(f"[dumps] Codec {codec} not supported")
    return DESERIALIZERS.get(codec)(o)



import os
import dotenv
import logging
import logging.config

# load .env file
dotenv.load_dotenv()

AIRTABLE_TOKEN = os.getenv("BAM_AIRTABLE_TOKEN", None)
AIRTABLE_BASE_ID = os.getenv("BAM_AIRTABLE_BASE_ID", None)
AIRTABLE_ASSISTANCE_REQUESTS_TABLE_NAME = os.getenv(
    "BAM_AIRTABLE_ASSISTANCE_REQUESTS_TABLE_NAME", "Assistance Requests: Main"
)
AIRTABLE_VOLUNTEERS_TABLE_NAME = os.getenv(
    "BAM_AIRTABLE_VOLUNTEERS_TABLE_NAME", "Volunteers: Main"
)
AIRTABLE_ESSENTIAL_GOODS_TABLE_NAME = os.getenv(
    "BAM_AIRTABLE_ESSENTIAL_GOODS_TABLE_NAME", "Essential Good Donations: Main"
)
AIRTABLE_MESH_VIEW_NAME = os.getenv(
    "BAM_AIRTABLE_MESH_VIEW_NAME", "MESH - Pending installs, by address (Lu)"
)

# logging settings
LOG_LEVEL = os.getenv("BAM_LOG_LEVEL", "INFO")
LOG_FORMAT = os.getenv(
    "BAM_LOG_FORMAT",
    r"<%(levelname)s> %(module)s:%(lineno)d -> %(message)s [%(asctime)s]",
)

LOGGING_CONFIG = {
    "version": 1,
    "disable_existing_loggers": True,
    "formatters": {"default": {"format": (LOG_FORMAT)}},
    "datefmt": "%I:%M:%S",
    "handlers": {
        "console": {
            "class": "logging.StreamHandler",
            "level": LOG_LEVEL,
            "formatter": "default",
        }
    },
    "root": {"level": LOG_LEVEL, "handlers": ["console"]},
}

# listmonk settings

LISTMONK_USERNAME = os.getenv("BAM_LISTMONK_USERNAME", None)
LISTMONK_PASSWORD = os.getenv("BAM_LISTMONK_PASSWORD", None)
LISTMONK_BASE_URL = os.getenv(
    "BAM_LISTMONK_BASE_URL", "https://lists.baml.ink/"
)
if not LISTMONK_BASE_URL.endswith("/"):
    LISTMONK_BASE_URL += "/"

# s3 settings
DO_TOKEN = os.getenv("BAM_DO_TOKEN", None)
S3_BASE_URL = os.getenv(
    "BAM_S3_BASE_URL", "https://nyc3.digitaloceanspaces.com"
)
S3_ENDPOINT_URL = os.getenv(
    "BAM_S3_ENDPOINT_URL", "https://nyc3.digitaloceanspaces.com"
)
S3_ACCESS_KEY_ID = os.getenv("BAM_S3_ACCESS_KEY_ID", None)
S3_SECRET_ACCESS_KEY = os.getenv("BAM_S3_SECRET_ACCESS_KEY", None)
S3_BUCKET = os.getenv("BAM_S3_BUCKET", "bam-file")
S3_REGION_NAME = os.getenv("BAM_S3_REGION_NAME", "nyc3")
S3_PLATFORM = os.getenv("BAM_S3_PLATFORM", "do")
S3_CDN_ID = os.getenv("BAM_S3_CDN_ID", None)

logging.config.dictConfig(LOGGING_CONFIG)



import os
from typing import List, NewType

Filepath = NewType("Filepath", str)  # A filepath string


def list_files(path: Filepath, ignore_hidden: bool = False) -> List[str]:
    f"""
    Recursively list files under a directory.
    :param path: A filepath as a string
    :param ignore_hidden: whether or not to ignore hidden files (starting with ``.``)
    :return list
    """
    return (
        os.path.join(dp, f)
        for dp, dn, fn in os.walk(get_full(path))
        for f in fn
        if not (f.startswith(".") and ignore_hidden)
        and not f.endswith(".DS_Store")
        and not dp == "__MACOSX"
    )


def get_full(path: Filepath) -> str:
    f"""
    Get a full path
    :param path: A filepath as a string
    :return str
    """
    path = os.path.expandvars(path)
    path = os.path.expanduser(path)
    path = os.path.normpath(path)
    path = os.path.abspath(path)
    return path



import os
from setuptools import setup, find_packages

from bam_core.__version__ import VERSION

reqs = os.path.abspath(
    os.path.join(os.path.dirname(__file__), "requirements.txt")
)
with open(reqs) as f:
    install_requires = [req.strip().split("==")[0] for req in f]

config = {
    "name": "bam-core",
    "version": VERSION,
    "packages": find_packages(),
    "install_requires": install_requires,
    "author": "BAM",
    "author_email": "hola@bushwickayudamutua.com",
    "description": "Shared code for BAM Automations",
    "url": "http://bushwickayudamutua.org",
}

setup(**config)



from bam_core.settings import *

def test_no_secrets_are_published():
    # ensure that secrets are not published to the repo in settings files
    assert S3_ACCESS_KEY_ID is None
    assert S3_SECRET_ACCESS_KEY is None
    assert DO_TOKEN is None
    assert LISTMONK_USERNAME is None
    assert LISTMONK_PASSWORD is None
    assert AIRTABLE_TOKEN is None
    assert AIRTABLE_BASE_ID is None
